#
# Copyright (c) 2008-2017 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'


# import standard library
from persistent import Persistent
try:
    from osgeo.osr import SpatialReference, CoordinateTransformation
    have_gdal = True
except ImportError:
    have_gdal = False

# import interfaces
from hypatia.interfaces import ICatalog
from onf_website.component.location.interfaces import IViewLocationSettings, VIEW_LOCATION_SETTINGS_KEY, ILocationInfo
from pyams_content.shared.view.interfaces import IWfView, IViewSettings, IViewQueryParamsExtension, \
    IViewQueryEsParamsExtension
from zope.annotation.interfaces import IAnnotations

# import packages
from elasticsearch_dsl import Q
from hypatia.query import Any
from pyams_catalog.query import and_
from pyams_utils.adapter import adapter_config, ContextAdapter
from pyams_utils.registry import get_utility
from pyramid.threadlocal import get_current_registry
from zope.container.contained import Contained
from zope.interface import implementer
from zope.lifecycleevent import ObjectCreatedEvent
from zope.location import locate
from zope.schema.fieldproperty import FieldProperty


@implementer(IViewLocationSettings)
class ViewLocationSettings(Persistent, Contained):
    """View location settings"""

    select_context_gps_location = FieldProperty(IViewLocationSettings['select_context_gps_location'])
    gps_location = FieldProperty(IViewLocationSettings['gps_location'])
    gps_distance = FieldProperty(IViewLocationSettings['gps_distance'])
    gps_area = FieldProperty(IViewLocationSettings['gps_area'])
    select_context_forests = FieldProperty(IViewLocationSettings['select_context_forests'])
    forests = FieldProperty(IViewLocationSettings['forests'])
    select_context_cities = FieldProperty(IViewLocationSettings['select_context_cities'])
    cities = FieldProperty(IViewLocationSettings['cities'])
    select_context_departments = FieldProperty(IViewLocationSettings['select_context_departments'])
    departments = FieldProperty(IViewLocationSettings['departments'])
    select_context_countries = FieldProperty(IViewLocationSettings['select_context_countries'])
    countries = FieldProperty(IViewLocationSettings['countries'])
    select_context_structures = FieldProperty(IViewLocationSettings['select_context_structures'])
    structures = FieldProperty(IViewLocationSettings['structures'])

    @property
    def is_using_context(self):
        return self.select_context_gps_location or self.select_context_forests or self.select_context_cities or \
               self.select_context_departments or self.select_context_countries or self.select_context_structures

    def get_gps_location(self, context):
        if self.select_context_gps_location:
            location = ILocationInfo(context, None)
            if location is not None:
                coords = location.gps_location.wgs_coordinates
                if coords and coords[0] and coords[1]:
                    return {'filter': 'geo_distance',
                            'location': {
                                "lon": coords[0],
                                "lat": coords[1]
                            },
                            'distance': '{0}km'.format(self.gps_distance)}
        else:
            if self.gps_location:
                coords = self.gps_location.wgs_coordinates
                if coords and coords[0] and coords[1]:
                    return {'filter': 'geo_distance',
                            'location': {
                                "lon": coords[0],
                                "lat": coords[1]
                            },
                            'distance': '{0}km'.format(self.gps_distance)}
            elif self.gps_area:
                coords = self.gps_area.wgs_coordinates
                if coords and coords[0] and coords[1]:
                    return {'filter': 'geo_bounding_box',
                            'area': {
                                "top_left": {
                                    "lon": coords[0][0],
                                    "lat": coords[1][1]
                                },
                                "bottom_right": {
                                    "lon": coords[1][0],
                                    "lat": coords[0][1]
                                }
                            }}

    def get_forests(self, context):
        forests = set()
        if self.select_context_forests:
            location = ILocationInfo(context, None)
            if location is not None:
                forests |= location.forests or set()
        if self.forests:
            forests |= set(self.forests)
        return list(forests)

    def get_cities(self, context):
        cities = set()
        if self.select_context_cities:
            location = ILocationInfo(context, None)
            if location is not None:
                cities |= location.cities or set()
        if self.cities:
            cities |= set(self.cities)
        return list(cities)

    def get_departments(self, context):
        departments = set()
        if self.select_context_departments:
            location = ILocationInfo(context, None)
            if location is not None:
                departments |= location.departments or set()
        if self.departments:
            departments |= self.departments
        return departments

    def get_departments_index(self, context):
        return [department.insee_code for department in self.get_departments(context)]

    def get_countries(self, context):
        countries = set()
        if self.select_context_countries:
            location = ILocationInfo(context, None)
            if location is not None:
                countries |= location.countries or set()
        if self.countries:
            countries |= self.countries
        return countries

    def get_countries_index(self, context):
        return [country.insee_code for country in self.get_countries(context)]

    def get_structures(self, context):
        structures = set()
        if self.select_context_structures:
            location = ILocationInfo(context, None)
            if location is not None:
                structures |= location.structures or set()
        if self.structures:
            structures |= set(self.structures)
        return list(structures)


@adapter_config(context=IWfView, provides=IViewLocationSettings)
@adapter_config(name='location', context=IWfView, provides=IViewSettings)
def ViewLocationSettingsFactory(view):
    """View location settings factory"""
    annotations = IAnnotations(view)
    settings = annotations.get(VIEW_LOCATION_SETTINGS_KEY)
    if settings is None:
        settings = annotations[VIEW_LOCATION_SETTINGS_KEY] = ViewLocationSettings()
        get_current_registry().notify(ObjectCreatedEvent(settings))
        locate(settings, view, '++view:location++')
    return settings


@adapter_config(name='location', context=IWfView, provides=IViewQueryParamsExtension)
class ViewLocationQueryParamsExtension(ContextAdapter):
    """View location query params extension"""

    weight = 50

    def get_params(self, context):
        catalog = get_utility(ICatalog)
        settings = IViewLocationSettings(self.context)
        params = None
        # check forests
        forests = settings.get_forests(context)
        if forests:
            params = and_(params, Any(catalog['forests'], forests))
        # check cities
        cities = settings.get_cities(context)
        if cities:
            params = and_(params, Any(catalog['cities'], cities))
        # check departments
        departments = settings.get_departments_index(context)
        if departments:
            params = and_(params, Any(catalog['departments'], departments))
        # check countries
        countries = settings.get_countries_index(context)
        if countries:
            params = and_(params, Any(catalog['countries'], countries))
        # structures
        structures = settings.get_structures(context)
        if structures:
            params = and_(params, Any(catalog['structures'], structures))
        return params


@adapter_config(name='location', context=IWfView, provides=IViewQueryEsParamsExtension)
class ViewLocationQueryEsParamsExtension(ContextAdapter):
    """View location query params extension for Elasticsearch"""

    weight = 50

    def get_es_params(self, context):
        settings = IViewLocationSettings(self.context)
        params = []
        # GPS location
        gps_filter = None
        location = settings.get_gps_location(context)
        if location:
            filter = location.pop('filter')
            filter_args = None
            if filter == 'geo_distance':
                filter_args = {'location.coords': location['location'],
                               'distance': location['distance']}
            elif filter == 'geo_bounding_box':
                filter_args = {'location.coords': location['area']}
            if filter_args:
                gps_filter = Q(filter, **filter_args)
        # forests
        forests = settings.get_forests(context)
        if forests:
            params.append(Q('terms', **{'location.forests': forests}))
        # cities
        cities = settings.get_cities(context)
        if cities:
            params.append(Q('terms', **{'location.cities': cities}))
        # check departments
        departments = settings.get_departments_index(context)
        if departments:
            params.append(Q('terms', **{'location.departments': departments}))
        # check countries
        countries = settings.get_countries_index(context)
        if countries:
            params.append(Q('terms', **{'location.countries': countries}))
        # structures
        structures = settings.get_structures(context)
        if structures:
            params.append(Q('terms', **{'location.structures': structures}))
        # check query arguments
        kwargs = {}
        if params:
            kwargs['must'] = params
        if gps_filter:
            kwargs['filter'] = gps_filter
        if kwargs:
            return Q('bool', **kwargs)
